use bitflags::bitflags;

use super::{InterruptStackFrame, VirtAddr};

bitflags! {
	/// Describes an page fault error code.
	///
	/// This structure is defined by the following manual sections:
	///   * AMD Volume 2: 8.4.2
	///   * Intel Volume 3A: 4.7
	#[repr(transparent)]
	#[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy)]
	pub struct PageFaultErrorCode: u64 {
		/// If this flag is set, the page fault was caused by a page-protection violation,
		/// else the page fault was caused by a not-present page.
		const PROTECTION_VIOLATION = 1;

		/// If this flag is set, the memory access that caused the page fault was a write.
		/// Else the access that caused the page fault is a memory read. This bit does not
		/// necessarily indicate the cause of the page fault was a read or write violation.
		const CAUSED_BY_WRITE = 1 << 1;

		/// If this flag is set, an access in user mode (CPL=3) caused the page fault. Else
		/// an access in supervisor mode (CPL=0, 1, or 2) caused the page fault. This bit
		/// does not necessarily indicate the cause of the page fault was a privilege violation.
		const USER_MODE = 1 << 2;

		/// If this flag is set, the page fault is a result of the processor reading a 1 from
		/// a reserved field within a page-translation-table entry.
		const MALFORMED_TABLE = 1 << 3;

		/// If this flag is set, it indicates that the access that caused the page fault was an
		/// instruction fetch.
		const INSTRUCTION_FETCH = 1 << 4;

		/// If this flag is set, it indicates that the page fault was caused by a protection key.
		const PROTECTION_KEY = 1 << 5;

		/// If this flag is set, it indicates that the page fault was caused by a shadow stack
		/// access.
		const SHADOW_STACK = 1 << 6;

		/// If this flag is set, it indicates that the page fault was caused by SGX access-control
		/// requirements (Intel-only).
		const SGX = 1 << 15;

		/// If this flag is set, it indicates that the page fault is a result of the processor
		/// encountering an RMP violation (AMD-only).
		const RMP = 1 << 31;
	}
}

//
// MARK: types
//

/// A handler function for an interrupt or an exception without error code.
pub type HandlerFunc = extern "x86-interrupt" fn(InterruptStackFrame);

/// A handler function for an exception that pushes an error code.
pub type HandlerFuncWithErrCode = extern "x86-interrupt" fn(InterruptStackFrame, error_code: u64);

/// A handler function that must not return, e.g. for a machine check exception.
pub type DivergingHandlerFunc = extern "x86-interrupt" fn(InterruptStackFrame) -> !;

/// A handler function with an error code that must not return, e.g. for a double fault exception.
pub type DivergingHandlerFuncWithErrCode =
	extern "x86-interrupt" fn(InterruptStackFrame, error_code: u64) -> !;

/// A page fault handler function that pushes a page fault error code.
pub type PageFaultHandlerFunc =
	extern "x86-interrupt" fn(InterruptStackFrame, error_code: PageFaultErrorCode);

/// A common trait for all handler functions usable in [`Entry`].
///
/// # Safety
///
/// Implementors have to ensure that `to_virt_addr` returns a valid address.
pub unsafe trait HandlerFuncType {
	/// Get the virtual address of the handler function.
	fn to_virt_addr(self) -> VirtAddr;
}

unsafe impl HandlerFuncType for HandlerFunc {
	#[inline]
	fn to_virt_addr(self) -> VirtAddr {
		#[allow(clippy::fn_to_numeric_cast_with_truncation)]
		VirtAddr::new(self as u32)
	}
}

unsafe impl HandlerFuncType for HandlerFuncWithErrCode {
	#[inline]
	fn to_virt_addr(self) -> VirtAddr {
		#[allow(clippy::fn_to_numeric_cast_with_truncation)]
		VirtAddr::new(self as u32)
	}
}

unsafe impl HandlerFuncType for DivergingHandlerFunc {
	#[inline]
	fn to_virt_addr(self) -> VirtAddr {
		#[allow(clippy::fn_to_numeric_cast_with_truncation)]
		VirtAddr::new(self as u32)
	}
}

unsafe impl HandlerFuncType for DivergingHandlerFuncWithErrCode {
	#[inline]
	fn to_virt_addr(self) -> VirtAddr {
		#[allow(clippy::fn_to_numeric_cast_with_truncation)]
		VirtAddr::new(self as u32)
	}
}

unsafe impl HandlerFuncType for PageFaultHandlerFunc {
	#[inline]
	fn to_virt_addr(self) -> VirtAddr {
		#[allow(clippy::fn_to_numeric_cast_with_truncation)]
		VirtAddr::new(self as u32)
	}
}
