#![no_std]
#![feature(int_roundings)]
#![feature(lang_items)]
#![feature(abi_x86_interrupt)]
#![allow(internal_features)]

use core::arch::asm;
use lazy_static::lazy_static;
use spin::Mutex;

use drivers::vga::Vga13h;
use idt::{InterruptDescriptorTable, InterruptStackFrame};
use os::thunk::ThunkData;
use tetrisboard::TetrisBoard;

mod idt;
mod os;
mod tetrisboard;

#[macro_use]
mod drivers;

pub(crate) static VGA: Mutex<Vga13h> = Mutex::new(unsafe { Vga13h::new() });

lazy_static! {
	static ref IDT: InterruptDescriptorTable = {
		let mut idt = InterruptDescriptorTable::new();
		idt.divide_error.set_handler_fn(breakpoint_handler);

		idt
	};
}

extern "x86-interrupt" fn breakpoint_handler(stack_frame: InterruptStackFrame) {
	println!("EXCEPTION {:?}", stack_frame);
}

#[no_mangle]
pub unsafe extern "C" fn start(
	_boot_disk: usize,
	thunk10: extern "C" fn(),
	_thunk13: extern "C" fn(),
	_thunk15: extern "C" fn(),
	_thunk16: extern "C" fn(),
) -> ! {
	println!("Entered Rust, serial ready.");

	{
		// Set vga mode
		let mut data = ThunkData::new();
		data.eax = 0x13;
		data.with(thunk10);
	}

	{
		// Disable cursor
		let mut data = ThunkData::new();
		data.eax = 0x0100;
		data.ecx = 0x3F00;
		data.with(thunk10);
	}

	{
		// Initialize OS
		IDT.load();
	}

	// Clear screen
	let mut v = VGA.lock();

	let t = TetrisBoard::new();
	t.draw(&mut v);

	asm!("int $0");

	panic!("kernel");
}
